# C 库封装

import sys
import os
import ctypes

# 当前文件所处的文件夹
__curdir = os.path.split(os.path.realpath(__file__))[0]
__dll = None

__dllPath = os.path.join(__curdir, "calc.dll")

if sys.platform == "win32":  # Windows
    try:
        __dll = ctypes.CDLL(__dllPath)
    except Exception as e:
        __dll = ctypes.WinDLL(__dllPath)

__jCalc = __dll.jCalc
__jCalc.argtypes = [ctypes.c_double * 13, ]
__jCalc.restype = ctypes.c_int
__mCalc = __dll.mCalc
__mCalc.argtypes = [ctypes.c_double * 13, ]
__mCalc.restype = ctypes.c_int


def jCalc(args):
    k = (13 * ctypes.c_double)()
    for i in range(13):
        k[i] = args[i]
    res = __jCalc(k)
    r = k[0]
    del k
    return res, r


def mCalc(args):
    k = (13 * ctypes.c_double)()
    for i in range(13):
        k[i] = args[i]
    res = __mCalc(k)
    r = k[0]
    del k
    return res, r



"""
    L-System for fractal
    已完成功能：
        1.简单D0L-系统
        2.合成D0L-系统
        3.简单分叉结构
    待完成功能：
        1.带年龄树（分叉）的生成
        2.随机L-系统
        3.参数L-系统
"""

from math import sin, cos, pi
import pygame
from .base import Base


def left(screen, st, angle, d):
    # st: 起点坐标
    # angle: 向左偏转的度数
    # d: 距离
    angle = pi * angle / 180
    return [st[0] + d * cos(angle), st[1] - d * sin(angle)]


class Pen(Base):

    def __init__(self, size, title=""):
        # size 画布的宽高 [width, hight]
        # title 画布标题
        Base.__init__(self, size, self.__run, title)
        self.setPoint([size[0] / 2, size[1] / 2])
        self.setColor([0, 0, 0])
        self.setWidth(1)
        self.setAngle(0)

    def setAngle(self, angle):
        self.angle = angle

    def setPoint(self, pos):
        # 设置笔的位置
        self.pos = pos

    def setWidth(self, width):
        # 设置线宽
        self.width = width

    def setColor(self, color):
        # 设置颜色
        self.color = color

    def left(self, angle):
        # 向左转angle度
        self.angle = self.angle + angle

    def right(self, angle):
        # 向右转angle度
        self.angle = self.angle - angle

    def forward(self, d):
        # 向前走d步长
        np = left(self.screen, self.pos, self.angle, d)
        pygame.draw.line(self.screen, self.color, self.pos, np, self.width)
        self.pos = np

    def __run(self):
        # 线程
        self.draw(self.omega, self.P, self.delta, self.length)

    def draw(self, omega, P, delta, length):
        i = 0
        while i < len(omega):
            if omega[i] == '+':
                self.left(delta)
            elif omega[i] == '-':
                self.right(delta)
            elif omega[i] == '[':
                k = 0
                st = i
                while i < len(omega):
                    if omega[i] == "[":
                        k += 1
                    elif omega[i] == "]":
                        k -= 1
                    if k == 0:
                        break
                    i += 1
                sub = omega[st + 1:i]
                curpoint = self.pos[:]
                curangle = self.angle
                self.draw(sub, P, delta, length)
                self.pos = curpoint
                self.angle = curangle
            else:
                self.forward(length)
            i += 1

    def doD0L(self, omega, P, delta, times, length, rate):
        # omega: 公理（初始字符串）
        # P: 产生式（映射规则）
        # delta: 角度增量
        # times: 迭代次数
        # length: 初始线长
        # rate: 每次迭代后缩小的倍数
        length /= (rate**times)
        for i in range(times):  # 完成字符串迭代
            ct = 0
            for key in P:
                omega = omega.replace(key, str(ct))
                ct += 1
            ct = 0
            for key in P:
                omega = omega.replace(str(ct), P[key])
                ct += 1
        self.omega = omega
        self.P = P
        self.delta = delta
        self.length = length



"""
    迭代函数系统（IFS）
"""
import pygame
from random import random
from math import sin, cos
from .base import Base


def getIfsCode(matrix):
    # 极坐标形式变换到矩阵
    res = []
    for i in matrix:
        r, s, the, fu, e, f, p = i
        item = [r * cos(the), -s * sin(fu), r * sin(the), s * cos(fu), e, f, p]
        res.append(item)
    return res


class IFS(Base):

    def __init__(self, size, title="", color=None):
        Base.__init__(self, size, self.__run, title)
        self.setPx()
        self.setIfsCode()
        self.__coo = True
        if color:
            self.color = color
        else:
            self.color = [255, 255, 255]
        self.screen.fill(self.color)

    def setCoordinate(self):
        # 纵轴的反向
        self.__coo = not self.__coo

    def ifsp(self, x, y):
        # 变换规则(返回迭代后的坐标)
        # 此函数具体用时需要重写
        return (x, y)

    def setIfsp(self, call):
        self.ifsp = call

    def setIfsCode(self, ifsCode=None):
        # 设置ifs码
        self.ifsCode = ifsCode
        if ifsCode != None:
            # 解析ifs码
            p = [i[-1] for i in self.ifsCode]  # 每个方程的概率
            self.__fn = len(p)  # 方程个数
            sp = sum(p)
            # 将概率映射到0-1区间
            self.__pe = [sum(p[:i + 1])/sp for i in range(self.__fn)]

    def __parseIfsCode(self, x, y):
        rand = random()
        i = 0
        while i < self.__fn:
            if rand <= self.__pe[i]:  # 选择概率pe[i]对应的方程
                a, b, c, d, e, f, p = self.ifsCode[i]
                return (a * x + b * y + e, c * x + d * y + f)
            i += 1
        return (0, 0)

    def setPx(self, enlarge=1, pl=0, pt=0):
        # 平移、放大调整
        # enlarge: 放大倍数
        # pl: 向右平移像素
        # pt: 向下平移像素
        self.enlarge = enlarge
        self.pl = pl
        self.pt = pt

    def __run(self):
        start = self.start
        for i in range(self.n):
            if self.__coo:
                self.screen.set_at(
                    (int(self.enlarge * start[0] + self.pl), int(self.enlarge * start[1] + self.pt)), self.pcolor)
            else:
                self.screen.set_at((int(self.enlarge * start[0] + self.pl), self.screen.get_height(
                ) - int(self.enlarge * start[1] + self.pt)), self.pcolor)
            start = self.ifsp(*start)

    def doIFS(self, n, start=None, color=None):
        """
            开始迭代
            start: 迭代起点
            color: 描点的颜色
        """
        self.n = n
        if start == None:
            self.start = (0, 0)
        else:
            self.start = start
        if color == None:
            self.pcolor = [0, 0, 0]
        else:
            self.pcolor = color
        if self.ifsCode != None:
            self.ifsp = self.__parseIfsCode



"""
    基类
    打印（保存图片）图片：Ctrl+P
    放大：点击放大部分鼠标左键
    缩小：点击缩小部分鼠标右键
"""

import pygame
from pygame.locals import K_p, K_LCTRL, K_RCTRL  # 按下Ctrl + P键打印
import threading
import time

class Base(object):

    def __init__(self, size, callRun, title=""):
        """
            size: 画布尺寸
            callRun: 子类的绘画函数回掉
            title: 画布标题
        """
        self.__run = callRun
        self.title = title
        pygame.init()
        self.screen = pygame.display.set_mode(size)
        pygame.display.set_caption(title)
        self.screen.fill([255, 255, 255])
        self.psave = False
        self.scalaRate = 4  # 放大（缩小）倍数
        self.ctrl = False  # Ctrl键按下
        self.cnt = 1  # 连续命名时标号

    def save(self, title):
        # 保存图片，title为文件名
        pygame.image.save(self.screen, title)

    def wait(self):
        # 等待用户关闭
        th = threading.Thread(target=self.__run)
        th.start()
        while 1:
            for event in pygame.event.get():
                if not th.is_alive():  # 必须绘图结束才可操作
                    if event.type == pygame.QUIT:
                        exit()
                    if event.type == pygame.KEYDOWN:
                        if event.key == K_p and self.ctrl:
                            if self.title == "":
                                self.title = "game"
                            self.save(self.title + str(self.cnt) + ".jpg")
                            self.ctrl = False
                            self.cnt += 1
                        if event.key == K_LCTRL or event.key == K_RCTRL:
                            self.ctrl = True
                    if event.type == pygame.MOUSEBUTTONDOWN:
                        if event.button == 1 and hasattr(self, "scala"):  # 鼠标左键放大
                            i, j = event.pos
                            w = self.screen.get_width() // (2 * self.scalaRate)
                            h = self.screen.get_height() // (2 * self.scalaRate)
                            pygame.draw.rect(self.screen, [0, 0, 0], [
                                             i - w, j - h, 2 * w, 2 * h], 1)
                            pygame.display.flip()
                            time.sleep(1)
                            self.scala(i, j, self.scalaRate)
                            th = threading.Thread(target=self.__run)
                            th.start()
                        # 鼠标右键缩小
                        elif event.button == 3 and hasattr(self, "scala"):
                            i, j = event.pos
                            self.scala(i, j, 1 / self.scalaRate)
                            th = threading.Thread(target=self.__run)
                            th.start()
            pygame.display.flip()



from setuptools import setup, find_packages

setup(
    name="fractal",
    version="0.2.7",
    keywords=["fractal", "分形"],
    description="对分形比较感兴趣，看pypi上没有相关库，自己撸着玩",
    license="MIT",
    author="陈粮",
    author_email="1570184051@qq.com",
    packages=find_packages(),
    install_requires=["pygame>=1.9.3"],
    platforms="any",
    url = "https://github.com/ChenL1994/fractal",
    zip_safe=False,
    package_data = {
        "": ["*.dll",]
    }
)



import pygame
from .colors import *
from .base import Base
from threading import Thread

calc = None

try:
    from .clib import mCalc
    calc = mCalc
    print("Loading C library ...")
except:
    pass

class Mandelbrot(Base):

    def __init__(self, size, title=""):
        Base.__init__(self, size, self.__run, title)
        self.setExp(2)
        self.setRadius(2)
        self.setZ0(0 + 0j)
        self.width = size[0]
        self.height = size[1]
        self.setRange(3.5, 3.5)
        self.setCentre(0 + 0j)

    def setRadius(self, R):
        # 设置逃逸半径
        self.R = R

    def setZ0(self, Z0):
        # 设置起始迭代复数（一般为0+0j）
        self.Z0 = Z0

    def setCentre(self, z0):
        # 设置中心点
        self.z0 = z0

    def setRange(self, xmax, ymax):
        # 设置坐标范围，范围越小图放大倍数越高
        self.xmax = xmax
        self.ymax = ymax

    def __getXY(self, i, j):
        # 通过像素坐标获取映射后的坐标
        return complex((i / self.width - 0.5) * self.xmax + self.z0.real, (j / self.height - 0.5) * self.ymax + self.z0.imag)

    def scala(self, i, j, rate):
        # 将(i, j)像素点置于中心位置，放大rete倍
        self.setCentre(self.__getXY(i, j))
        self.xmax /= rate
        self.ymax /= rate

    def setExp(self, expc):
        # 设置指数，默认2
        self.expc = expc

    def color(self, i, r = 2):
        # 对第i次迭代点着色，返回RGB值
        if i < len(reds) - 1:
            return reds[i]
        elif r < self.R:
            j = (len(blues) - 1) * r / self.R
            return blues[-int(j)]
        return (0, 0, 0)

    def setColor(self, call):
        self.color = call

    def __calc(self, start, w, h):
        # 绘制以start为起点，宽w,高h的子区域
        for i in range(w):
            for j in range(h):
                if calc:  # 如果加载动态链接库没问题
                    ct, r = mCalc([start[0] + i, start[1] + j, self.Z0.real, self.Z0.imag, self.z0.real,
                                self.z0.imag, self.width, self.height, self.xmax, self.ymax, self.N, self.expc, self.R])
                    self.screen.set_at(
                        [start[0] + i, start[1] + j], self.color(ct, r))
                else:
                    ct = 0
                    z = self.Z0
                    c = self.__getXY(start[0] + i, start[1] + j)
                    for k in range(self.N):
                        ct = k
                        if abs(z) > self.R:  # 大于逃逸半径，则返回
                            break
                        z = z**self.expc + c
                    self.screen.set_at(
                        [start[0] + i, start[1] + j], self.color(ct, abs(z)))

    def __run(self):
        # 绘图
        print("x range ：[-%f,%f]\ny range ：[-%f,%f]" % (
            self.xmax, self.xmax, self.ymax, self.ymax))
        tn = 5  # 25 个子线程绘图
        ci = self.width // tn
        cj = self.height // tn
        ts = []
        for i in range(tn):
            for j in range(tn):
                t = Thread(target=self.__calc, args=(
                    [i * ci, j * cj], self.width // tn, self.height // tn))
                t.start()
                ts.append(t)
        for t in ts:
            t.join()
        del ts

    def doMandelbrot(self, N):
        # 开始迭代
        # N: 最大迭代次数
        self.N = N
